import hashlib

def generate_node_id(public_key_bytes):
  # Use SHA-1 hash of the public key as the node ID
  node_id = hashlib.sha1(public_key_bytes).hexdigest()
  return node_id

def xor_distance(id1, id2):
  return int(id1, 16) ^ int(id2, 16)



# File: cli.py
import os
import subprocess
import sys
import shutil
import tempfile
import re
from pathlib import Path
from importlib import resources  # Python 3.9+

def read_existing_env(env_path):
    existing_vars = {}
    if env_path.exists():
        with open(env_path, 'r') as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith('#'):
                    key, value = line.split('=', 1)
                    existing_vars[key.strip()] = value.strip()
    return existing_vars

def validate_port(port):
    try:
        port = int(port)
        return 1024 <= port <= 65535
    except ValueError:
        return False

def validate_hostname(hostname):
    if not hostname:
        return False
    ip_pattern = re.compile(r'^(\d{1,3}\.){3}\d{1,3}|[a-zA-Z0-9\.-]+$')
    return bool(ip_pattern.match(hostname))

def prompt_for_three_vars(existing_vars):
    """Prompt ONLY for NODE_PORT, NODE_TCP_PORT, ADVERTISE_IP.
    If user hits Enter, keep existing .env value or use defaults."""
    
    # 1) NODE_PORT
    default_node_port = existing_vars.get('NODE_PORT', '8000')
    while True:
        node_port = input(
            f"Enter NODE_PORT (press Enter for default {default_node_port}): "
        ).strip()
        if not node_port:
            node_port = default_node_port
        if validate_port(node_port):
            existing_vars["NODE_PORT"] = node_port
            break
        print("Invalid port! Must be between 1024 and 65535.")

    # 2) NODE_TCP_PORT
    default_tcp_port = existing_vars.get('NODE_TCP_PORT', '8500')
    while True:
        node_tcp_port = input(
            f"Enter NODE_TCP_PORT (press Enter for default {default_tcp_port}): "
        ).strip()
        if not node_tcp_port:
            node_tcp_port = default_tcp_port
        if validate_port(node_tcp_port):
            existing_vars["NODE_TCP_PORT"] = node_tcp_port
            break
        print("Invalid port! Must be between 1024 and 65535.")

    # 3) ADVERTISE_IP
    default_ip = existing_vars.get('ADVERTISE_IP', 'localhost')
    while True:
        advertise_ip = input(
            f"Enter ADVERTISE_IP (e.g., example.com) [default {default_ip}]: "
        ).strip()
        if not advertise_ip:
            advertise_ip = default_ip
        if validate_hostname(advertise_ip):
            existing_vars["ADVERTISE_IP"] = advertise_ip
            break
        print("Invalid hostname/IP! Please enter a valid hostname or IP address.")

def write_env(env_path, vars_dict):
    try:
        with open(env_path, 'w') as f:
            for k, v in vars_dict.items():
                f.write(f"{k}={v}\n")
    except Exception as e:
        print(f"Error writing .env: {e}")
        sys.exit(1)

def up():
    print("\nSetting up environment variables...")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        # Copy Dockerfile + docker-compose.yml from inside the package
        docker_dir = resources.files("sokoweb.docker")
        shutil.copyfile(docker_dir / "Dockerfile", temp_path / "Dockerfile")
        shutil.copyfile(docker_dir / "docker-compose.yml", temp_path / "docker-compose.yml")

        # Now handle the .env. If user already has .env in their project,
        # you might want to let them specify the path. Or we can start empty.
        env_path = temp_path / ".env"

        # If you want to load an existing .env from the user's current directory,
        # you can copy that in if it exists:
        user_env = Path.cwd() / ".env"
        if user_env.exists():
            shutil.copyfile(user_env, env_path)

        existing_vars = read_existing_env(env_path)
        # Prompt ONLY for the 3 required variables (others remain from .env)
        prompt_for_three_vars(existing_vars)

        # Write them back
        write_env(env_path, existing_vars)

        print("\nUpdated environment variables (from .env in the temp directory):")
        for k, v in existing_vars.items():
            print(f"{k}={v}")

        # Start Docker
        print("\nStarting Docker containers...")
        try:
            process = subprocess.run(
                ["docker", "compose", "-f", "docker-compose.yml", "up", "--build"],
                check=True,
                cwd=str(temp_path)
            )
            if process.returncode == 0:
                print("Successfully started Docker containers.")
        except subprocess.CalledProcessError as e:
            print(f"Error starting Docker containers (exit code={e.returncode})")
            sys.exit(e.returncode)
        except Exception as e:
            print(f"An unexpected error occurred: {str(e)}")
            sys.exit(1)

def down():
    """Stop/remove containers. For simplicity, runs a global 'docker compose down -v'."""
    print("Stopping Docker containers and removing volumes...")
    try:
        subprocess.run(["docker", "compose","-f", str(temp_path / "docker-compose.yml"),"down", "-v"], check=True)
        print("Successfully stopped and removed containers/volumes.")
    except subprocess.CalledProcessError as e:
        print(f"Error stopping Docker containers (exit code={e.returncode})")
        sys.exit(e.returncode)
    except Exception as e:
        print(f"An unexpected error occurred: {str(e)}")
        sys.exit(1)

if __name__ == "__main__":
    up()


# crypto.py
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives import serialization, hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
import os
import logging

logger = logging.getLogger(__name__)

def generate_key_pair():
  private_key = ed25519.Ed25519PrivateKey.generate()
  public_key = private_key.public_key()
  return private_key, public_key

def sign_message(private_key, message_bytes):
    """
    Sign the message bytes using the provided private key.

    :param private_key: The private key object for signing.
    :param message_bytes: The message bytes to sign.
    :return: Signature bytes.
    """
    signature = private_key.sign(message_bytes)
    return signature  # Return bytes



def verify_signature(public_key_bytes, message_bytes, signature_bytes):
    public_key = ed25519.Ed25519PublicKey.from_public_bytes(public_key_bytes)
    try:
        public_key.verify(signature_bytes, message_bytes)
        return True
    except InvalidSignature:
        return False



def serialize_public_key(public_key):
  return public_key.public_bytes(
      encoding=serialization.Encoding.Raw,
      format=serialization.PublicFormat.Raw
  )

def serialize_private_key(private_key):
  return private_key.private_bytes(
      encoding=serialization.Encoding.Raw,
      format=serialization.PrivateFormat.Raw,
      encryption_algorithm=serialization.NoEncryption()
  )

# Symmetric encryption functions

def derive_key(password: bytes, salt: bytes):
  """
  Derive a symmetric key from a password and salt using PBKDF2.

  :param password: The password bytes.
  :param salt: The salt bytes.
  :return: Derived key bytes.
  """
  kdf = PBKDF2HMAC(
      algorithm=hashes.SHA256(),
      length=32,  # AES-256 key length
      salt=salt,
      iterations=100000,
      backend=default_backend()
  )
  key = kdf.derive(password)
  return key

def encrypt_data(key: bytes, data: bytes):
  """
  Encrypt data using AES CBC mode.

  :param key: Symmetric key bytes.
  :param data: Data bytes to encrypt.
  :return: Tuple of (IV, ciphertext).
  """
  iv = os.urandom(16)
  cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend())
  encryptor = cipher.encryptor()
  # Pad data to block size
  padder = padding.PKCS7(128).padder()
  padded_data = padder.update(data) + padder.finalize()
  ciphertext = encryptor.update(padded_data) + encryptor.finalize()
  return iv + ciphertext  # Prepend IV for use in decryption

def decrypt_data(key: bytes, iv_ciphertext: bytes):
  """
  Decrypt data using AES CBC mode.

  :param key: Symmetric key bytes.
  :param iv_ciphertext: Concatenated IV and ciphertext.
  :return: Decrypted data bytes.
  """
  iv = iv_ciphertext[:16]
  ciphertext = iv_ciphertext[16:]
  cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend())
  decryptor = cipher.decryptor()
  decrypted_padded = decryptor.update(ciphertext) + decryptor.finalize()
  # Remove padding
  unpadder = padding.PKCS7(128).unpadder()
  data = unpadder.update(decrypted_padded) + unpadder.finalize()
  return data

def encrypt_phone_number(phone_number: str, password: str) -> bytes:
    """
    Encrypts the phone number using the provided password.

    :param phone_number: The phone number string to encrypt.
    :param password: The password string to derive the key.
    :return: The encrypted data as bytes (salt + iv + ciphertext).
    """
    # Convert strings to bytes
    phone_data = phone_number.encode("utf-8")
    password_bytes = password.encode("utf-8")

    # Generate a random salt
    salt = os.urandom(16)  # 16 bytes salt

    # Derive key from password and salt
    key = derive_key(password_bytes, salt)

    # Encrypt data
    iv_ciphertext = encrypt_data(key, phone_data)

    # Return the concatenation of salt + iv + ciphertext
    return salt + iv_ciphertext

def decrypt_phone_number(encrypted_data: bytes, password: str) -> str:
    """
    Decrypts the encrypted phone number using the provided password.

    :param encrypted_data: The encrypted data bytes (salt + iv + ciphertext).
    :param password: The password string to derive the key.
    :return: The decrypted phone number string.
    """
    # Convert password to bytes
    password_bytes = password.encode("utf-8")

    # Extract the salt
    salt = encrypted_data[:16]  # First 16 bytes
    iv_ciphertext = encrypted_data[16:]

    # Derive key from password and salt
    key = derive_key(password_bytes, salt)

    # Decrypt data
    decrypted_data = decrypt_data(key, iv_ciphertext)

    # Convert bytes to string
    phone_number = decrypted_data.decode("utf-8")
    return phone_number



# storage_messages.py

import json
import logging

logger = logging.getLogger(__name__)

class StorageMessage:
    def __init__(self, message_type, sender_id, data, signature=None, public_key=None):
        self.message_type = message_type
        self.sender_id = sender_id
        self.data = data
        self.signature = signature
        self.public_key = public_key

    def to_json_unsigned(self):
        d = {
            "message_type": self.message_type,
            "sender_id": self.sender_id,
            "data": self.data,
        }
        json_str = json.dumps(d, sort_keys=True)
        logger.debug(f"Serialized unsigned storage message to JSON: {json_str}")
        return json_str

    def to_json(self):
        d = {
            "message_type": self.message_type,
            "sender_id": self.sender_id,
            "data": self.data,
            "signature": self.signature,
            "public_key": self.public_key,
        }
        json_str = json.dumps(d, sort_keys=True)
        logger.debug(f"Serialized storage message to JSON: {json_str}")
        return json_str

    @staticmethod
    def from_json(json_str):
        logger.debug(f"Deserializing storage message from JSON: {json_str}")
        obj = json.loads(json_str)
        return StorageMessage(
            message_type=obj["message_type"],
            sender_id=obj["sender_id"],
            data=obj["data"],
            signature=obj.get("signature"),
            public_key=obj.get("public_key"),
        )


# db_models.py

from sqlalchemy import Column, String, Float, Boolean, LargeBinary, JSON
from sqlalchemy.dialects.postgresql import UUID
import uuid

from .database import Base


class User(Base):
  __tablename__ = 'users'

  id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
  username = Column(String, unique=True, index=True, nullable=False)
  email = Column(String, nullable=True)
  full_name = Column(String, nullable=True)
  hashed_password = Column(String, nullable=False)
  disabled = Column(Boolean, default=False)
  scopes = Column(String, nullable=True)  # Comma-separated string of scopes
  credits = Column(Float, default=0.0)
  phone_number = Column(String, nullable=True)  # Added phone_number field

  # Relationships can be added if needed, e.g., products


class Product(Base):
    __tablename__ = "products"

    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    product_id = Column(String, unique=True, index=True, nullable=False)
    # Core attributes
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)
    category = Column(String, nullable=False)
    price = Column(Float, nullable=False)
    image_refs = Column(JSON, nullable=True)  # Store as JSON
    seller_location = Column(JSON, nullable=True)  # Store as JSON
    shop_name = Column(String, nullable=True)
    seller_phone_encrypted = Column(LargeBinary, nullable=True)
    # Extended attributes
    extended_attributes = Column(JSON, nullable=True)


  # Relationships can be added if needed

class NodeCredit(Base):
    __tablename__ = 'node_credits'

    node_id = Column(String, primary_key=True)
    credits = Column(Float, default=0.0)


# storage_service.py

import asyncio
import hashlib
import json
import logging
import time
import os
import struct
import aiofiles
import aiofiles.os as aioos

class StorageService:
    def __init__(
        self,
        node,
        storage_dir="storage_chunks",
        cleanup_interval=60,
        republish_interval=3600,
    ):
        self.node = node
        self.loop = asyncio.get_event_loop()
        self.cleanup_task = None
        self.republish_task = None
        self.logger = logging.getLogger(__name__)

        self.storage_dir = storage_dir
        self.chunks_dir = os.path.join(self.storage_dir, "chunks")
        self.sub_chunks_dir = os.path.join(self.storage_dir, "sub_chunks")

        os.makedirs(self.chunks_dir, exist_ok=True)
        os.makedirs(self.sub_chunks_dir, exist_ok=True)

        self.cleanup_interval = cleanup_interval
        self.republish_interval = republish_interval

        self.chunk_store = {}
        self.metadata_file = os.path.join(self.storage_dir, "chunk_store_meta.json")
        self._init_task = asyncio.create_task(self.load_chunk_store())

        self.logger.info(f"Initialized StorageService for node {self.node.node_id}")

    async def wait_for_initialization(self):
        await self._init_task

    async def start(self):
        self.cleanup_task = asyncio.create_task(
            self.cleanup_expired_chunks(self.cleanup_interval)
        )
        self.republish_task = asyncio.create_task(
            self.republish_chunks(self.republish_interval)
        )
        self.logger.info(f"StorageService started periodic tasks for {self.node.node_id}")

    async def stop(self):
        if self.cleanup_task:
            self.cleanup_task.cancel()
            try:
                await self.cleanup_task
            except asyncio.CancelledError:
                pass
            self.logger.info(f"StorageService cleanup task stopped for {self.node.node_id}")

        if self.republish_task:
            self.republish_task.cancel()
            try:
                await self.republish_task
            except asyncio.CancelledError:
                pass
            self.logger.info(f"StorageService republish task stopped for {self.node.node_id}")

        await self.save_chunk_store()
        self.logger.info(f"StorageService chunk_store saved for {self.node.node_id}")

    async def save_chunk_store(self):
        async with aiofiles.open(self.metadata_file, "w") as f:
            chunk_store_data = {k: str(v) for k, v in self.chunk_store.items()}
            await f.write(json.dumps(chunk_store_data))
        self.logger.debug(f"Saved chunk_store metadata: {chunk_store_data}")

    async def load_chunk_store(self):
        if await aioos.path.exists(self.metadata_file):
            async with aiofiles.open(self.metadata_file, "r") as f:
                content = await f.read()
                chunk_store_data = json.loads(content)
                self.chunk_store = {k: float(v) for k, v in chunk_store_data.items()}
            self.logger.debug(f"Loaded chunk_store metadata: {chunk_store_data}")
        else:
            self.chunk_store = {}

    async def distribute_file_to_network(self, file_hash, file_data, ttl):
        try:
            k = self.node.routing_table.k
            closest_nodes = await self.node.find_nodes(file_hash)
            if not closest_nodes:
                self.logger.warning(
                    f"{self.node.node_id}: No nodes available to store file {file_hash}"
                )
                return
            for n in closest_nodes:
                if n.node_id == self.node.node_id:
                    continue
                try:
                    await self.send_file_over_tcp(n, file_hash, file_data, ttl)
                except Exception as e:
                    self.logger.error(
                        f"Failed to send file {file_hash} to node {n.node_id} over TCP: {e}"
                    )
        except Exception as e:
            self.logger.error(
                f"Error distributing file {file_hash} to network: {e}", exc_info=True
            )

    async def handle_store_file(self, message, reader, writer):
        file_hash = message.get("file_hash")
        ttl = message.get("ttl", 604800)
        self.logger.debug(f"Received STORE_FILE message: {message}")

        if not file_hash:
            self.logger.error("Invalid STORE_FILE message received")
            if not writer.is_closing():
                writer.close()
            return
        try:
            file_size_data = await reader.readexactly(8)
            file_size = int.from_bytes(file_size_data, byteorder="big")
            self.logger.debug(f"Expecting {file_size} bytes of file data.")

            file_data = await reader.readexactly(file_size)
            self.logger.debug(f"Read {len(file_data)} bytes of file data.")

            calculated_hash = hashlib.sha256(file_data).hexdigest()
            if calculated_hash != file_hash:
                self.logger.error(
                    f"File hash mismatch: expected {file_hash}, got {calculated_hash}"
                )
                response = {"status": "ERROR", "error": "File hash mismatch"}
                rsp = json.dumps(response).encode("utf-8")
                l = len(rsp).to_bytes(4, byteorder="big")
                writer.write(l + rsp)
                await writer.drain()
                return

            await self.save_chunk_to_file(file_hash, file_data)

            expiration_time = time.time() + ttl
            self.chunk_store[file_hash] = expiration_time
            self.logger.info(f"Stored file {file_hash} of size {file_size} bytes")

            response = {"status": "OK"}
            rsp = json.dumps(response).encode("utf-8")
            l = len(rsp).to_bytes(4, byteorder="big")
            writer.write(l + rsp)
            await writer.drain()

        except Exception as e:
            self.logger.error(f"Failed to store file {file_hash}: {e}", exc_info=True)
            response = {"status": "ERROR", "error": str(e)}
            rsp = json.dumps(response).encode("utf-8")
            l = len(rsp).to_bytes(4, byteorder="big")
            try:
                writer.write(l + rsp)
                await writer.drain()
            except Exception as we:
                self.logger.error(f"Error sending error response for {file_hash}: {we}", exc_info=True)
            finally:
                if not writer.is_closing():
                    writer.close()

    async def handle_tcp_connection(self, reader, writer, message_json):
        addr = writer.get_extra_info("peername")
        self.logger.debug(f"Handling TCP from {addr} in StorageService")

        try:
            message = json.loads(message_json)
            message_type = message.get("message_type")
            if message_type == "STORE_FILE":
                await self.handle_store_file(message, reader, writer)
            elif message_type == "RETRIEVE_FILE":
                await self.handle_retrieve_file_tcp(message, reader, writer)
            else:
                self.logger.warning(f"Unknown storage message type over TCP: {message_type}")
        except (asyncio.CancelledError, GeneratorExit):
            self.logger.warning(
                f"TCP connection from {addr} was cancelled or encountered GeneratorExit."
            )
            raise
        except Exception as e:
            self.logger.error(f"Error handling TCP connection from {addr}: {e}", exc_info=True)
            response = {"status": "ERROR", "message": str(e)}
            rsp = json.dumps(response).encode("utf-8")
            try:
                writer.write(len(rsp).to_bytes(4, byteorder="big") + rsp)
                await writer.drain()
            except Exception as we:
                self.logger.error(f"Error sending error response to {addr}: {we}", exc_info=True)
        finally:
            if not writer.is_closing():
                writer.close()
                await writer.wait_closed()

    async def handle_retrieve_file_tcp(self, message, reader, writer):
        file_hash = message.get("file_hash")
        if file_hash:
            file_path = os.path.join(self.storage_dir, file_hash)
            if await aioos.path.exists(file_path):
                async with aiofiles.open(file_path, "rb") as f:
                    file_data = await f.read()
                response = {"status": "OK", "file_size": len(file_data)}
                rsp = json.dumps(response).encode("utf-8")
                length = len(rsp).to_bytes(4, byteorder="big")
                writer.write(length + rsp)
                await writer.drain()

                writer.write(file_data)
                await writer.drain()
            else:
                self.logger.warning(f"Requested file {file_hash} not found.")
                response = {"status": "ERROR", "message": "File not found"}
                rsp = json.dumps(response).encode("utf-8")
                length = len(rsp).to_bytes(4, byteorder="big")
                writer.write(length + rsp)
                await writer.drain()
        else:
            self.logger.warning("Invalid RETRIEVE_FILE message over TCP.")
            response = {"status": "ERROR", "message": "Invalid request"}
            rsp = json.dumps(response).encode("utf-8")
            length = len(rsp).to_bytes(4, byteorder="big")
            writer.write(length + rsp)
            await writer.drain()

    async def handle_store_chunk_tcp(self, message, reader, writer):
        chunk_hash = message.get("chunk_hash")
        chunk_size = message.get("chunk_size")
        ttl = message.get("ttl", 604800)
        if chunk_hash and chunk_size:
            data = await reader.readexactly(chunk_size)
            calc_hash = hashlib.sha256(data).hexdigest()
            if calc_hash == chunk_hash:
                await self.save_chunk_to_file(chunk_hash, data)
                expiration_time = time.time() + ttl
                self.chunk_store[chunk_hash] = expiration_time
                self.logger.info(f"Stored chunk {chunk_hash} from TCP connection")
                response = {"status": "OK"}
                rsp = json.dumps(response).encode("utf-8")
                writer.write(len(rsp).to_bytes(4, byteorder="big") + rsp)
                await writer.drain()
            else:
                self.logger.warning(f"Chunk hash mismatch for {chunk_hash} over TCP")
                response = {"status": "ERROR", "message": "Hash mismatch"}
                rsp = json.dumps(response).encode("utf-8")
                writer.write(len(rsp).to_bytes(4, byteorder="big") + rsp)
                await writer.drain()
        else:
            self.logger.warning("Invalid STORE_CHUNK message over TCP")
            response = {"status": "ERROR", "message": "Invalid request"}
            rsp = json.dumps(response).encode("utf-8")
            writer.write(len(rsp).to_bytes(4, byteorder="big") + rsp)
            await writer.drain()

    async def handle_retrieve_chunk_tcp(self, message, reader, writer):
        chunk_hash = message.get("chunk_hash")
        if (
            chunk_hash in self.chunk_store
            and time.time() < self.chunk_store[chunk_hash]
        ):
            data = await self.read_chunk_from_file(chunk_hash)
            if data:
                response = {"status": "OK", "chunk_size": len(data)}
                rsp = json.dumps(response).encode("utf-8")
                writer.write(len(rsp).to_bytes(4, byteorder="big") + rsp)
                await writer.drain()

                writer.write(data)
                await writer.drain()
            else:
                response = {"status": "ERROR", "message": "Chunk data not found"}
                rsp = json.dumps(response).encode("utf-8")
                writer.write(len(rsp).to_bytes(4, byteorder="big") + rsp)
                await writer.drain()
        else:
            response = {"status": "ERROR", "message": "Chunk not available"}
            rsp = json.dumps(response).encode("utf-8")
            writer.write(len(rsp).to_bytes(4, byteorder="big") + rsp)
            await writer.drain()

    async def save_chunk_to_file(self, file_hash: str, data: bytes) -> None:
        file_path = os.path.join(self.storage_dir, file_hash)
        async with aiofiles.open(file_path, "wb") as f:
            await f.write(data)
        self.logger.debug(f"Saved file {file_hash} to local storage.")

    async def read_chunk_from_file(self, file_hash):
        file_path = os.path.join(self.chunks_dir, file_hash)
        if await aioos.path.exists(file_path):
            async with aiofiles.open(file_path, "rb") as f:
                data = await f.read()
            self.logger.debug(f"Read chunk {file_hash} from {file_path}")
            return data
        else:
            self.logger.debug(f"Chunk file {file_path} does not exist")
            return None

    async def cleanup_expired_chunks(self, interval=60):
        while True:
            try:
                await asyncio.sleep(interval)
                current_time = time.time()
                expired = [
                    ch for ch, exp in self.chunk_store.items() if current_time >= exp
                ]
                for ch in expired:
                    await self.delete_chunk(ch)
            except asyncio.CancelledError:
                break
            except Exception as e:
                self.logger.error(f"Error in cleanup_expired_chunks: {e}")

    async def delete_chunk(self, chunk_hash):
        if chunk_hash in self.chunk_store:
            del self.chunk_store[chunk_hash]
        file_path = os.path.join(self.chunks_dir, chunk_hash)
        if await aioos.path.exists(file_path):
            await aioos.remove(file_path)
            self.logger.debug(f"Deleted chunk file {chunk_hash}")

    async def republish_chunks(self, interval=3600):
        while True:
            try:
                await asyncio.sleep(interval)
                self.logger.debug("Running republish_chunks task")
                current_time = time.time()
                for ch, exp_time in self.chunk_store.items():
                    if current_time < exp_time:
                        data = await self.read_chunk_from_file(ch)
                        if data:
                            remaining_ttl = exp_time - current_time
                            await self.store_chunk(ch, data, ttl=remaining_ttl)
            except asyncio.CancelledError:
                break
            except Exception as e:
                self.logger.error(f"Error in republish_chunks: {e}")

    async def store_chunk(self, chunk_hash: str, chunk_data: bytes, ttl=604800) -> None:
        await self.save_chunk_to_file(chunk_hash, chunk_data)
        expiration_time = time.time() + ttl
        self.chunk_store[chunk_hash] = expiration_time

        closest_nodes = await self.node.find_nodes(chunk_hash)
        if not closest_nodes:
            self.logger.warning(
                f"{self.node.node_id}: No nodes available to store chunk {chunk_hash}"
            )
            return

        for n in closest_nodes:
            try:
                await self.send_chunk_over_tcp(n, chunk_hash, chunk_data, ttl)
            except Exception as e:
                self.logger.error(
                    f"Failed to send chunk {chunk_hash} to node {n.node_id} over TCP: {e}"
                )

    async def send_chunk_over_tcp(self, node, chunk_hash, chunk_data, ttl):
        reader, writer = await asyncio.open_connection(node.ip, node.tcp_port)
        try:
            message = {
                "message_type": "STORE_CHUNK",
                "chunk_hash": chunk_hash,
                "chunk_size": len(chunk_data),
                "ttl": ttl,
            }
            msg_json = json.dumps(message).encode("utf-8")
            writer.write(len(msg_json).to_bytes(4, byteorder="big") + msg_json)
            await writer.drain()
            writer.write(chunk_data)
            await writer.drain()

            data = await reader.readexactly(4)
            rsp_len = int.from_bytes(data, byteorder="big")
            rsp_data = await reader.readexactly(rsp_len)
            rsp = json.loads(rsp_data.decode("utf-8"))
            if rsp.get("status") != "OK":
                raise Exception(f"Failed to store chunk on node {node.node_id}: {rsp.get('message')}")
        finally:
            writer.close()
            await writer.wait_closed()

    async def retrieve_chunk_from_network(self, chunk_hash):
        closest_nodes = await self.node.find_nodes(chunk_hash)
        for n in closest_nodes:
            try:
                chunk_data = await self.request_chunk_over_tcp(n, chunk_hash)
                if chunk_data:
                    if hashlib.sha256(chunk_data).hexdigest() == chunk_hash:
                        return chunk_data
                    else:
                        self.logger.warning(
                            f"Received corrupt chunk from node {n.node_id}"
                        )
            except Exception as e:
                self.logger.error(
                    f"Error retrieving chunk from node {n.node_id}: {e}"
                )
        return None

    async def request_chunk_over_tcp(self, node, chunk_hash):
        reader, writer = await asyncio.open_connection(node.ip, node.tcp_port)
        try:
            message = {"message_type": "RETRIEVE_CHUNK", "chunk_hash": chunk_hash}
            msg_json = json.dumps(message).encode("utf-8")
            writer.write(len(msg_json).to_bytes(4, byteorder="big") + msg_json)
            await writer.drain()

            data = await reader.readexactly(4)
            rsp_len = int.from_bytes(data, byteorder="big")
            rsp_data = await reader.readexactly(rsp_len)
            rsp = json.loads(rsp_data.decode("utf-8"))
            if rsp.get("status") == "OK":
                csize = rsp.get("chunk_size")
                cdata = await reader.readexactly(csize)
                return cdata
            else:
                self.logger.warning(
                    f"Failed to retrieve chunk {chunk_hash} from node {node.node_id}: {rsp.get('message')}"
                )
                return None
        finally:
            writer.close()
            await writer.wait_closed()

    async def store_file(self, file_hash: str, file_data: bytes, ttl=604800) -> None:
        await self.save_chunk_to_file(file_hash, file_data)
        expiration_time = time.time() + ttl
        self.chunk_store[file_hash] = expiration_time

        closest_nodes = await self.node.find_nodes(file_hash)
        if not closest_nodes:
            self.logger.warning(
                f"{self.node.node_id}: No nodes available to store file {file_hash}"
            )
            return

        for n in closest_nodes:
            if n.node_id == self.node.node_id:
                continue
            success = await self.send_file_over_tcp(n, file_hash, file_data, ttl)
            if not success:
                self.logger.error(
                    f"Failed to send file {file_hash} to node {n.node_id} over TCP"
                )

    async def send_file_over_tcp(self, node, file_hash, file_data, ttl):
        try:
            reader, writer = await asyncio.open_connection(node.ip, node.tcp_port)
            message = {
                "message_type": "STORE_FILE",
                "file_hash": file_hash,
                "file_size": len(file_data),
                "ttl": ttl,
            }
            msg = json.dumps(message).encode("utf-8")
            msg_len = len(msg).to_bytes(4, byteorder="big")
            writer.write(msg_len + msg)
            await writer.drain()

            file_len = len(file_data).to_bytes(8, byteorder="big")
            writer.write(file_len + file_data)
            await writer.drain()

            try:
                resp_len_data = await asyncio.wait_for(reader.readexactly(4), timeout=30)
                resp_len = int.from_bytes(resp_len_data, byteorder="big")
                resp_data = await reader.readexactly(resp_len)
                resp = json.loads(resp_data.decode("utf-8"))
                if resp.get("status") == "OK":
                    self.logger.info(f"File {file_hash} sent successfully to {node.node_id}")
                    return True
                else:
                    self.logger.error(
                        f"Failed to send file {file_hash} to node {node.node_id}: {resp.get('error', 'Unknown error')}"
                    )
                    return False
            except asyncio.TimeoutError:
                self.logger.error(f"Timeout waiting for ack from node {node.node_id}")
                return False
            except asyncio.IncompleteReadError as e:
                self.logger.error(f"Incomplete read waiting for ack from {node.node_id}: {e}")
                return False
        except Exception as e:
            self.logger.error(f"Error sending file to node {node.node_id}: {e}", exc_info=True)
            return False
        finally:
            writer.close()
            try:
                await writer.wait_closed()
            except:
                pass

    async def retrieve_file_from_network(self, file_hash: str) -> bytes:
        file_path = os.path.join(self.storage_dir, file_hash)
        try:
            if await aioos.path.exists(file_path):
                self.logger.debug(f"File {file_hash} found locally.")
                async with aiofiles.open(file_path, "rb") as f:
                    return await f.read()
        except Exception as e:
            self.logger.error(f"Error reading local file {file_hash}: {e}")

        closest_nodes = await self.node.find_nodes(file_hash)
        if not closest_nodes:
            self.logger.warning(f"No nodes found close to file hash {file_hash}.")
            return None

        for n in closest_nodes:
            try:
                self.logger.debug(
                    f"Attempting to retrieve file {file_hash} from node {n.node_id}."
                )
                file_data = await self.request_file_over_tcp(n, file_hash)
                if file_data:
                    calc_hash = hashlib.sha256(file_data).hexdigest()
                    if calc_hash != file_hash:
                        raise Exception(
                            f"File hash mismatch: expected {file_hash}, got {calc_hash}"
                        )
                    await self.save_chunk_to_file(file_hash, file_data)
                    self.chunk_store[file_hash] = time.time() + 604800
                    self.logger.debug(
                        f"Retrieved and stored file {file_hash} from node {n.node_id}."
                    )
                    return file_data
            except Exception as e:
                self.logger.error(f"Failed to retrieve file {file_hash} from {n.node_id}: {e}")
        self.logger.warning(f"Failed to retrieve file {file_hash} from any node.")
        return None

    async def request_file_over_tcp(self, node, file_hash):
        try:
            reader, writer = await asyncio.open_connection(node.ip, node.tcp_port)
            try:
                msg = {
                    "message_type": "RETRIEVE_FILE",
                    "file_hash": file_hash,
                }
                msg_json = json.dumps(msg).encode("utf-8")
                writer.write(len(msg_json).to_bytes(4, byteorder="big") + msg_json)
                await writer.drain()

                resp_len_data = await reader.readexactly(4)
                resp_len = int.from_bytes(resp_len_data, byteorder="big")
                resp_data = await reader.readexactly(resp_len)
                resp = json.loads(resp_data.decode("utf-8"))

                if resp.get("status") == "OK":
                    file_size = resp.get("file_size")
                    if file_size is None:
                        raise Exception("Missing file_size in response")
                    file_size = int(file_size)
                    self.logger.debug(
                        f"Expecting {file_size} bytes from node {node.node_id}"
                    )
                    file_data = await reader.readexactly(file_size)
                    self.logger.debug(
                        f"Read {len(file_data)} bytes of file data from node {node.node_id}"
                    )
                    calc_hash = hashlib.sha256(file_data).hexdigest()
                    if calc_hash != file_hash:
                        raise Exception(
                            f"File hash mismatch: expected {file_hash}, got {calc_hash}"
                        )
                    return file_data
                else:
                    self.logger.warning(
                        f"Node {node.node_id} responded with error: {resp.get('message')}"
                    )
                    return None
            finally:
                writer.close()
                await writer.wait_closed()
        except Exception as e:
            self.logger.error(
                f"Error requesting file {file_hash} from node {node.node_id}: {e}"
            )
            return None